# -*- coding: utf-8 -*-
# Copyright (c) 2022-present tandemdude
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import abc
import enum
import typing as t

__all__ = [
    "TokenType",
    "Token",
    "LiteralMixin",
    "OperatorToken",
    "IdentifierToken",
    "StringToken",
    "IntToken",
    "FloatToken",
    "ErrorToken",
]


class TokenType(enum.Enum):
    # Compound tokens
    IDENTIFIER = "\\w[\\w\\d]+"
    INT_LITERAL = "[\\d]+"
    FLOAT_LITERAL = "[\\d]+\\.[\\d]*"
    STRING_LITERAL = "'.*'"
    # Special tokens
    UNKNOWN = "UNKNOWN"
    # Operators
    OPERATOR_EQ = "=="
    OPERATOR_NOT_EQ = "!="
    OPERATOR_GREATER_THAN = ">"
    OPERATOR_LESS_THAN = "<"
    OPERATOR_GREATER_EQUALS = ">="
    OPERATOR_LESS_EQUALS = "<="
    OPERATOR_NOT = "!"
    OPERATOR_AND = "&&"
    OPERATOR_OR = "||"
    OPERATOR_ADD = "+"
    OPERATOR_SUB = "-"
    OPERATOR_MULTIPLY = "*"
    OPERATOR_DIVIDE = "/"
    OPERATOR_INTEGER_DIVIDE = "//"
    OPERATOR_MODULO = "%"
    OPERATOR_POW = "**"
    # Special characters
    QUESTION_MARK = "?"
    COLON = ":"
    ACCESSOR = "."
    COMMA = ","
    L_PAREN = "("
    R_PAREN = ")"
    L_SQUARE_BRACKET = "["
    R_SQUARE_BRACKET = "]"


EXCLUDED_TOKEN_TYPES = {
    TokenType.IDENTIFIER,
    TokenType.INT_LITERAL,
    TokenType.FLOAT_LITERAL,
    TokenType.STRING_LITERAL,
    TokenType.UNKNOWN,
}
SORTED_TOKEN_TYPES = sorted(
    set(TokenType).difference(EXCLUDED_TOKEN_TYPES),
    key=lambda itm: len(itm.value),
    reverse=True,
)


class Token(abc.ABC):
    __slots__ = ("value", "at", "type")

    value: t.Any
    at: int
    type: TokenType

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(value={self.value!r}, type={self.type})"


class LiteralMixin:
    __slots__ = ()


class OperatorToken(Token):
    __slots__ = ()

    def __init__(self, type: TokenType, at: int) -> None:
        self.value = type.value
        self.at = at
        self.type = type


class IdentifierToken(Token):
    __slots__ = ()

    type: TokenType = TokenType.IDENTIFIER

    def __init__(self, value: str, at: int) -> None:
        self.value = value
        self.at = at


class StringToken(Token, LiteralMixin):
    __slots__ = ()

    type: TokenType = TokenType.STRING_LITERAL

    def __init__(self, value: str, at: int) -> None:
        self.value = value.replace("\\", "")
        self.at = at


class IntToken(Token, LiteralMixin):
    __slots__ = ()

    type: TokenType = TokenType.INT_LITERAL

    def __init__(self, value: str, at: int) -> None:
        self.value: int = int(value)
        self.at = at


class FloatToken(Token, LiteralMixin):
    __slots__ = ()

    type: TokenType = TokenType.FLOAT_LITERAL

    def __init__(self, value: str, at: int) -> None:
        self.value = float(value)
        self.at = at


class ErrorToken(Token):
    __slots__ = ()

    type: TokenType = TokenType.UNKNOWN

    def __init__(self, value: str, at: int) -> None:
        self.value: str = value
        self.at = at



# -*- coding: utf-8 -*-
# Copyright (c) 2022-present tandemdude
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import typing as t

from pysel import errors
from pysel import tokens

__all__ = ["Lexer"]


class Lexer:
    __slots__ = ("raw", "tokens", "idx", "eof")

    def __init__(self, raw: str) -> None:
        self.raw: str = raw
        self.tokens: t.List[tokens.Token] = []

        self.idx: int = -1
        self.eof: bool = False

    def next(self) -> t.Tuple[t.Optional[str], t.Optional[int]]:
        self.idx += 1

        if len(self.raw) <= self.idx:
            return None, None

        if self.raw[self.idx].isspace():
            return self.next()

        return self.raw[self.idx], self.idx

    def peek(self, n_ahead: int = 1) -> t.Optional[str]:
        idx = self.idx + n_ahead

        if len(self.raw) <= idx:
            return None

        return self.raw[idx]

    def peek_to_length(self, n_ahead: int = 1) -> str:
        return self.raw[self.idx : self.idx + n_ahead]

    def parse_while(
        self,
        char: t.Optional[str],
        condition: t.Callable[[t.Optional[str], int, t.List[str]], bool],
        fail_on_eof: bool = False,
    ) -> str:
        buf: t.List[str] = []
        offset: int = 0
        while condition(char, offset, buf):
            if fail_on_eof and char is None:
                raise errors.ExpressionSyntaxError("Unexpected EOF while parsing", self.raw, [self.idx])
            assert char is not None
            buf.append(char)
            char, offset = self.peek(offset + 1), offset + 1
        return "".join(buf)

    def as_token(self, char: str, charindex: int) -> t.Tuple[tokens.Token, int]:
        if char.isalpha() or char == "_":
            tkn = self.parse_while(char, lambda c, o, _: c is not None and (c.isalnum() or c == "_"))
            return tokens.IdentifierToken(tkn, charindex), len(tkn)
        if char == "'" or char == '"':
            looking_for = "'" if char == "'" else '"'

            tkn = self.parse_while(
                "",
                lambda c, o, _: c != looking_for or self.raw[self.idx + o - 1 : self.idx + o + 1] == f"\\{looking_for}",
                True,
            )
            return tokens.StringToken(tkn, charindex), len(tkn) + 2
        if char.isdecimal():
            tkn = self.parse_while(
                char,
                lambda c, _, b: c is not None and (c.isdecimal() or (c == "." and "." not in b)),
            )
            if "." in tkn:
                return tokens.FloatToken(tkn, charindex), len(tkn)
            return tokens.IntToken(tkn, charindex), len(tkn)

        for token_type in tokens.SORTED_TOKEN_TYPES:
            value: str = token_type.value
            if len(value) == 1:
                if char != value:
                    continue
                return tokens.OperatorToken(token_type, charindex), 1

            if (tkn := self.peek_to_length(len(value))) == value:
                return tokens.OperatorToken(token_type, charindex), len(tkn)

        return tokens.ErrorToken(char, charindex), 1

    def tokenize(self) -> t.List[tokens.Token]:
        while not self.eof:
            char, charindex = self.next()

            if char is None:
                self.eof = True
                break

            assert charindex is not None
            token, consumed = self.as_token(char, charindex)
            if consumed > 1:
                self.idx += consumed - 1

            self.tokens.append(token)

        if invalid := list(filter(lambda tk: isinstance(tk, tokens.ErrorToken), self.tokens)):
            raise errors.ExpressionSyntaxError(
                "Unexpected tokens encountered during lexing",
                self.raw,
                [i.at for i in invalid],
            )

        return self.tokens



# -*- coding: utf-8 -*-
# Copyright (c) 2022-present tandemdude
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import typing as t

from pysel import ast
from pysel import errors
from pysel import lexer
from pysel import tokens

__all__ = ["ast", "lexer", "tokens", "Expression"]

__version__ = "0.0.2"

T = t.TypeVar("T")


class Expression(t.Generic[T]):
    __slots__ = ("raw", "ast")

    def __init__(self, raw: str) -> None:
        self.raw: str = raw
        self.ast: t.Optional[ast.Node] = None

    def to_ast(self) -> ast.Node:
        if self.ast is not None:
            return self.ast
        self.ast = ast.Parser(self.raw, lexer.Lexer(self.raw).tokenize()).compilation_unit()
        return self.ast

    def evaluate(self, env: t.Optional[t.Dict[str, t.Any]] = None) -> T:
        env = env or {}
        for primitive in (bool, float, int, str):
            env.setdefault(primitive.__name__, primitive)
        return t.cast(T, self.to_ast().evaluate(env or {}))



# -*- coding: utf-8 -*-
# Copyright (c) 2022-present tandemdude
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import typing as t

__all__ = ["ExpressionSyntaxError"]


class ExpressionSyntaxError(SyntaxError):
    def __init__(self, message: str, expr: t.Optional[str], at: t.Sequence[int]) -> None:
        built_message = [message]
        if expr:
            built_message.append((" " * 4) + repr(expr))
            if at:
                error_arrows = [" " for _ in range(5 + len(expr))]
                for idx in at:
                    error_arrows[5 + idx] = "^"
                built_message.append("".join(error_arrows))
        super().__init__("\n".join(built_message))



# -*- coding: utf-8 -*-
# Copyright (c) 2022-present tandemdude
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import annotations

import abc
import collections
import operator as operator_
import typing as t

from pysel import errors
from pysel import tokens as tokens_

__all__ = [
    "Node",
    "Literal",
    "Reference",
    "UnaryOp",
    "BinaryOp",
    "TernaryOp",
    "Accessor",
    "MethodCall",
    "Parser",
]

T = t.TypeVar("T")
UNARY_OPERATOR_MAPPING = {
    "!": operator_.not_,
    "-": operator_.neg,
    "+": operator_.pos,
}
BINARY_OPERATOR_MAPPING = {
    "==": operator_.eq,
    "!=": operator_.ne,
    ">": operator_.gt,
    "<": operator_.lt,
    ">=": operator_.ge,
    "<=": operator_.le,
    "+": operator_.add,
    "-": operator_.sub,
    "*": operator_.mul,
    "/": operator_.truediv,
    "//": operator_.floordiv,
    "%": operator_.mod,
    "**": operator_.pow,
}


class Node(abc.ABC):
    __slots__ = ()

    @abc.abstractmethod
    def evaluate(self, env: t.Mapping[str, t.Any]) -> t.Any:
        ...


class Literal(Node):
    __slots__ = ("value",)

    def __init__(self, value: T) -> None:
        self.value: T = value

    def __repr__(self) -> str:
        return f"Literal({self.value!r})"

    def evaluate(self, env: t.Mapping[str, t.Any]) -> T:
        return self.value


class Reference(Node):
    __slots__ = ("name",)

    def __init__(self, name: str) -> None:
        self.name = name

    def __repr__(self) -> str:
        return f"Reference({self.name!r})"

    def evaluate(self, env: t.Mapping[str, t.Any]) -> t.Any:
        return env[self.name]


class UnaryOp(Node):
    __slots__ = ("operator", "operand")

    def __init__(self, operator: str, operand: Node) -> None:
        self.operator = operator
        self.operand = operand

    def __repr__(self) -> str:
        return f"UnaryOp({self.operator}, {self.operand})"

    def evaluate(self, env: t.Mapping[str, t.Any]) -> t.Any:
        return UNARY_OPERATOR_MAPPING[self.operator](self.operand.evaluate(env))  # type: ignore[operator]


class BinaryOp(Node):
    __slots__ = ("lh", "operator", "rh")

    def __init__(self, lh: Node, operator: str, rh: Node) -> None:
        self.lh = lh
        self.operator = operator
        self.rh = rh

    def __repr__(self) -> str:
        return f"BinaryOp({self.lh}, {self.operator}, {self.rh})"

    def evaluate(self, env: t.Mapping[str, t.Any]) -> t.Any:
        # We process 'or' and 'and' separately to allow operator short-circuiting
        if self.operator == "||":
            return self.lh.evaluate(env) or self.rh.evaluate(env)
        elif self.operator == "&&":
            return self.lh.evaluate(env) and self.rh.evaluate(env)

        return BINARY_OPERATOR_MAPPING[self.operator](self.lh.evaluate(env), self.rh.evaluate(env))


class TernaryOp(Node):
    __slots__ = ("condition", "when_true", "when_false")

    def __init__(self, condition: Node, when_true: Node, when_false: Node) -> None:
        self.condition = condition
        self.when_true = when_true
        self.when_false = when_false

    def __repr__(self) -> str:
        return f"TernaryOp({self.condition} ? {self.when_true} : {self.when_false})"

    def evaluate(self, env: t.Mapping[str, t.Any]) -> t.Any:
        if self.condition.evaluate(env):
            return self.when_true.evaluate(env)
        return self.when_false.evaluate(env)


class Accessor(Node):
    __slots__ = ("operand", "attr")

    def __init__(self, operand: Node, attr: str) -> None:
        self.operand = operand
        self.attr = attr

    def __repr__(self) -> str:
        return f"Accessor({self.operand}, {self.attr!r})"

    def evaluate(self, env: t.Mapping[str, t.Any]) -> t.Any:
        return getattr(self.operand.evaluate(env), self.attr)


class MethodCall(Node):
    __slots__ = ("operand", "arguments")

    def __init__(self, operand: Node, arguments: t.Sequence[Node]) -> None:
        self.operand = operand
        self.arguments = arguments

    def __repr__(self) -> str:
        return f"MethodCall({self.operand}, args={self.arguments})"

    def evaluate(self, env: t.Mapping[str, t.Any]) -> t.Any:
        return self.operand.evaluate(env)(*(a.evaluate(env) for a in self.arguments))


class Getitem(Node):
    __slots__ = ("operand", "params")

    def __init__(self, operand: Node, params: t.Sequence[t.Optional[Node]]) -> None:
        self.operand = operand
        self.params: t.List[t.Optional[Node]] = list(params)

    def __repr__(self) -> str:
        return f"Getitem({self.operand}, params={self.params})"

    def evaluate(self, env: t.Mapping[str, t.Any]) -> t.Any:
        if len(self.params) > 1:
            self.params.extend([None] * (3 - len(self.params)))
            return self.operand.evaluate(env)[slice(*(p.evaluate(env) if p else None for p in self.params))]
        assert len(self.params) == 1 and self.params[0] is not None
        return self.operand.evaluate(env)[self.params[0].evaluate(env)]


class Parser:
    __slots__ = ("raw", "tokens", "idx", "error_stack")

    def __init__(self, raw: str, tokens: t.List[tokens_.Token]) -> None:
        self.raw = raw
        self.tokens = tokens

        self.idx = -1
        self.error_stack: collections.deque[str] = collections.deque()

    def syntax_error(self) -> t.NoReturn:
        char_indexes: t.List[int] = []

        if self.idx == -1:
            char_indexes.append(0)
        elif self.idx >= len(self.tokens):
            char_indexes.append(len(self.raw) - 1)
        else:
            curr_token = self.tokens[self.idx]
            char_indexes.extend(range(curr_token.at, curr_token.at + len(str(curr_token.value))))

        if self.error_stack:
            raise errors.ExpressionSyntaxError(
                f"Expected {self.error_stack.popleft()!r} was not found",
                self.raw,
                char_indexes,
            )

        next_token = self.tokens[self.idx + 1 :][0]
        char_indexes = [*range(next_token.at, next_token.at + len(str(next_token.value)))]
        raise errors.ExpressionSyntaxError("Unexpected token encountered while parsing", self.raw, char_indexes)

    def next_token(self) -> t.Optional[tokens_.Token]:
        self.idx += 1

        if len(self.tokens) <= self.idx:
            return None

        return self.tokens[self.idx]

    def peek_next_token(self) -> t.Optional[tokens_.Token]:
        token = self.next_token()
        self.idx -= 1
        return token

    def val(self) -> Node:
        if self.peek_next_token() is None:
            self.syntax_error()
        elif self.peek_next_token().value == "(":  # type: ignore[union-attr]
            self.next_token()
            self.error_stack.appendleft(")")
            node = self.ternary()
            if not getattr(self.peek_next_token(), "value", None) == ")":
                self.syntax_error()
            self.error_stack.popleft()
            self.next_token()
            return node
        elif isinstance(self.peek_next_token(), tokens_.IdentifierToken):
            return Reference(self.next_token().value)  # type: ignore[union-attr]
        elif isinstance(self.peek_next_token(), tokens_.LiteralMixin):
            return Literal(self.next_token().value)  # type: ignore[union-attr]
        self.syntax_error()

    def accessor(self, initial_node: t.Optional[Node] = None) -> Node:
        node = initial_node or self.val()

        while (nxt := self.peek_next_token()) is not None and nxt.value == ".":
            self.error_stack.appendleft("identifier")
            self.next_token()

            if self.peek_next_token() is None:
                self.syntax_error()

            node = Accessor(node, self.next_token().value)  # type: ignore[union-attr]
            self.error_stack.popleft()

        return node

    def getitem(self, initial_node: t.Optional[Node] = None) -> Node:
        node = self.accessor(initial_node)

        while (nxt := self.peek_next_token()) is not None and nxt.value == "[":
            self.error_stack.appendleft("]")
            self.next_token()
            if (tkn := self.peek_next_token()) is None or tkn.value == "]":
                self.error_stack.appendleft("expr | :")
                self.syntax_error()

            params: t.List[t.Optional[Node]] = [
                self.ternary() if getattr(self.peek_next_token(), "value", None) != ":" else None
            ]
            while (nxt2 := self.peek_next_token()) is not None and nxt2.value == ":":
                if len(params) >= 3:
                    self.syntax_error()

                self.error_stack.appendleft("expr | :")
                self.next_token()
                if (tkn := self.peek_next_token()) is not None and tkn.value == ":":
                    self.error_stack.popleft()
                    params.append(None)
                    continue

                params.append(
                    self.ternary() if getattr(self.peek_next_token(), "value", None) not in (":", "]") else None
                )
                self.error_stack.popleft()

            if (tkn := self.peek_next_token()) is None or tkn.value != "]":
                self.syntax_error()

            node = Getitem(node, params)
            self.next_token()  # Remove R_SQUARE_BRACKET
            self.error_stack.popleft()

            if (nxt := self.peek_next_token()) is not None and nxt.value == ".":
                # Chained call special case
                node = self.accessor(node)

        return node

    def method_call(self) -> Node:
        node = self.getitem()

        while (nxt := self.peek_next_token()) is not None and nxt.value == "(":
            self.next_token()
            self.error_stack.appendleft(")")
            if (nxt := self.peek_next_token()) is None:
                self.syntax_error()
            if nxt.value == ")":
                node = MethodCall(node, [])
            else:
                arguments = [self.ternary()]
                while (nxt3 := self.peek_next_token()) is not None and nxt3.value == ",":
                    self.next_token()
                    self.error_stack.appendleft("expr")
                    arguments.append(self.ternary())
                    self.error_stack.popleft()

                if (tkn := self.peek_next_token()) is None or tkn.value != ")":
                    self.syntax_error()

                node = MethodCall(node, arguments)
            self.next_token()  # Remove R_PAREN
            self.error_stack.popleft()

            if (nxt := self.peek_next_token()) is not None and nxt.value in ["[", "."]:
                # Chained call special case
                node = self.getitem(node)

        return node

    def pow(self) -> Node:
        node = self.method_call()

        while (nxt := self.peek_next_token()) is not None and nxt.value == "**":
            op = self.next_token().value  # type: ignore[union-attr]
            self.error_stack.appendleft("expr")
            node = BinaryOp(self.method_call(), op, node)
            self.error_stack.popleft()

        return node

    def unary_expr(self) -> Node:
        node = None
        while (nxt := self.peek_next_token()) is not None and nxt.value in ("-", "+"):
            self.next_token()
            self.error_stack.appendleft("expr")
            node = UnaryOp(nxt.value, self.pow())
            self.error_stack.popleft()

        return node or self.pow()

    def multi(self) -> Node:
        node = self.unary_expr()

        while (nxt := self.peek_next_token()) is not None and nxt.value in (
            "*",
            "/",
            "//",
            "%",
        ):
            self.error_stack.appendleft("expr")
            node = BinaryOp(node, self.next_token().value, self.unary_expr())  # type: ignore[union-attr]
            self.error_stack.popleft()

        return node

    def expr(self) -> Node:
        node = self.multi()

        while (nxt := self.peek_next_token()) is not None and nxt.value in ("+", "-"):
            self.error_stack.appendleft("expr")
            node = BinaryOp(node, self.next_token().value, self.multi())  # type: ignore[union-attr]
            self.error_stack.popleft()

        return node

    def comparison(self) -> Node:
        node = self.expr()

        while (nxt := self.peek_next_token()) is not None and nxt.value in (
            "==",
            "!=",
            ">",
            "<",
            ">=",
            "<=",
        ):
            self.error_stack.appendleft("expr")
            node = BinaryOp(node, self.next_token().value, self.expr())  # type: ignore[union-attr]
            self.error_stack.popleft()

        return node

    def logical_not(self) -> Node:
        node = None
        while (nxt := self.peek_next_token()) is not None and nxt.value == "!":
            self.next_token()
            self.error_stack.appendleft("expr")
            node = UnaryOp(nxt.value, self.comparison())
            self.error_stack.popleft()

        return node or self.comparison()

    def logical_and(self) -> Node:
        node = self.logical_not()

        while (nxt := self.peek_next_token()) is not None and nxt.value == "&&":
            self.error_stack.appendleft("expr")
            node = BinaryOp(node, self.next_token().value, self.logical_not())  # type: ignore[union-attr]
            self.error_stack.popleft()

        return node

    def logical_or(self) -> Node:
        node = self.logical_and()

        while (nxt := self.peek_next_token()) is not None and nxt.value == "||":
            self.error_stack.appendleft("expr")
            node = BinaryOp(node, self.next_token().value, self.logical_and())  # type: ignore[union-attr]
            self.error_stack.popleft()

        return node

    def ternary(self) -> Node:
        node = self.logical_or()

        while (nxt := self.peek_next_token()) is not None and nxt.value == "?":
            self.next_token()
            self.error_stack.appendleft("expr")
            when_true = self.ternary()

            self.error_stack.appendleft(":")
            if (nxt2 := self.peek_next_token()) is not None and nxt2.value == ":":
                self.next_token()
                self.error_stack.popleft()
                node = TernaryOp(node, when_true, self.ternary())
            else:
                self.syntax_error()
            self.error_stack.popleft()

        return node

    def compilation_unit(self) -> Node:
        node = self.ternary()

        if self.tokens[self.idx + 1 :]:
            self.syntax_error()

        return node



# -*- coding: utf-8 -*-
# Copyright (c) 2022-present tandemdude
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os
import re
import types

from setuptools import find_namespace_packages
from setuptools import setup

name = "pysel"


def parse_meta():
    with open(os.path.join(name, "__init__.py")) as fp:
        code = fp.read()

    token_pattern = re.compile(
        r"^__(?P<key>\w+)?__\s*=\s*(?P<quote>(?:'{3}|\"{3}|'|\"))(?P<value>.*?)(?P=quote)",
        re.M,
    )

    groups = {}

    for match in token_pattern.finditer(code):
        group = match.groupdict()
        groups[group["key"]] = group["value"]

    return types.SimpleNamespace(**groups)


def long_description():
    with open("README.md") as fp:
        return fp.read()


def parse_requirements_file(path):
    with open(path) as fp:
        dependencies = (d.strip() for d in fp.read().split("\n") if d.strip())
        return [d for d in dependencies if not d.startswith("#")]


meta = parse_meta()

setup(
    name="pysel-lang",
    version=meta.version,
    description="A simple but powerful embedded expression language for Python",
    long_description=long_description(),
    long_description_content_type="text/markdown",
    author="tandemdude",
    author_email="tandemdude1@gmail.com",
    url="https://github.com/tandemdude/pysel",
    packages=find_namespace_packages(include=[name + "*"]),
    license="MIT",
    include_package_data=True,
    python_requires=">=3.8.0,<3.11",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Interpreters",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
